@@ -14,6 +14,7 @@ import threading
 import time
 from collections import namedtuple
 from concurrent.futures import ThreadPoolExecutor
+from contextlib import asynccontextmanager
 from datetime import datetime
 from decimal import Decimal
 from functools import cached_property
@@ -107,7 +108,8 @@ class Repository:
         self._concurrent = concurrent
         self._quiet = quiet
         self._slots = asyncio.Queue(maxsize=concurrent)
-        # We need actual integers for TQDM slot management
+        # We need actual integers for TQDM slot management in CLI, but this queue
+        # can also act as a semaphore in the general case
         for slot in range(2, concurrent + 2):
             self._slots.put_nowait(slot)
 
@@ -122,17 +124,38 @@ class Repository:
             self._executor = ThreadPoolExecutor(max_workers=self._concurrent)
             return self._executor
 
-    def _as_coroutine(self, func, *args, **kwargs):
-        if inspect.iscoroutinefunction(func):
+    def _awrap(self, func, *args, **kwargs):
+        if inspect.iscoroutinefunction(func) or inspect.isasyncgenfunction(func):
             return func(*args, **kwargs)
         else:
             loop = asyncio.get_running_loop()
             return loop.run_in_executor(self.executor, func, *args, **kwargs)
 
+    async def _aiter(self, func, *args, **kwargs):
+        unknown = self._awrap(func, *args, **kwargs)
+        if inspect.isawaitable(unknown):
+            result = await unknown
+        else:
+            result = unknown
+
+        if not isinstance(result, collections.abc.AsyncIterable):
+            result = utils.async_gen_wrapper(result)
+
+        async for value in result:
+            yield value
+
     @property
     def _unlocked(self):
         return hasattr(self, 'props')
 
+    @asynccontextmanager
+    async def _acquire_slot(self):
+        slot = await self._slots.get()
+        try:
+            yield slot
+        finally:
+            self._slots.put_nowait(slot)
+
     def default_serialization_hook(self, data, /):
         return utils.type_hint(data)
 
@@ -376,13 +399,17 @@ class Repository:
         else:
             key = None
 
-        print(ef.bold + 'Uploading config' + ef.rs)
-        await self._as_coroutine(self.backend.upload, 'config', self.serialize(config))
+        async with self._acquire_slot():
+            print(ef.bold + 'Uploading config' + ef.rs)
+            await self._awrap(self.backend.upload, 'config', self.serialize(config))
+
         self.props = props
         return utils.DefaultNamespace(config=config, key=key)
 
     async def _load_config(self):
-        data = await self._as_coroutine(self.backend.download, 'config')
+        async with self._acquire_slot():
+            data = await self._awrap(self.backend.download, 'config')
+
         return RepositoryProps(**self._instantiate_config(self.deserialize(data)))
 
     async def unlock(self, *, password=None, key=None):
@@ -474,12 +501,9 @@ class Repository:
         encrypted_snapshots = {}
 
         async def _download_snapshot(path, digest):
-            slot = await self._slots.get()
-            try:
+            async with self._acquire_slot():
                 logger.info('Downloading %s', path)
-                contents = await self._as_coroutine(self.backend.download, path)
-            finally:
-                self._slots.put_nowait(slot)
+                contents = await self._awrap(self.backend.download, path)
 
             logger.info('Verifying %s', path)
             if self.props.hash_digest(contents) != digest:
@@ -491,9 +515,7 @@ class Repository:
 
             encrypted_snapshots[path] = contents
 
-        for path in await self._as_coroutine(
-            self.backend.list_files, self.SNAPSHOT_PREFIX
-        ):
+        async for path in self._aiter(self.backend.list_files, self.SNAPSHOT_PREFIX):
             name, tag = self.parse_snapshot_location(path)
             digest = bytes.fromhex(name)
 
@@ -857,22 +879,21 @@ class Repository:
                     await asyncio.sleep(queue_timeout)
                     continue
 
-                slot = await self._slots.get()
-                try:
-                    exists = await self._as_coroutine(
-                        self.backend.exists, chunk.location
-                    )
-                    logger.info(
-                        'Processing chunk #%d, exists=%s, I=%d, L=%s',
-                        chunk.counter,
-                        exists,
-                        chunk.index,
-                        chunk.location,
-                    )
-                    if exists:
-                        _chunk_done(chunk)
-                        state.bytes_reused += chunk.stream_end - chunk.stream_start
-                    else:
+                async with self._acquire_slot():
+                    exists = await self._awrap(self.backend.exists, chunk.location)
+
+                logger.info(
+                    'Processing chunk #%d, exists=%s, I=%d, L=%s',
+                    chunk.counter,
+                    exists,
+                    chunk.index,
+                    chunk.location,
+                )
+                if exists:
+                    _chunk_done(chunk)
+                    state.bytes_reused += chunk.stream_end - chunk.stream_start
+                else:
+                    async with self._acquire_slot() as slot:
                         length = len(chunk.contents)
                         stream = io.BytesIO(chunk.contents)
                         iowrapper = utils.tqdmio(
@@ -884,16 +905,14 @@ class Repository:
                             rate_limiter=rate_limiter,
                         )
                         with stream, iowrapper:
-                            await self._as_coroutine(
+                            await self._awrap(
                                 self.backend.upload_stream,
                                 chunk.location,
                                 iowrapper,
                                 length,
                             )
 
-                        _chunk_done(chunk)
-                finally:
-                    self._slots.put_nowait(slot)
+                    _chunk_done(chunk)
 
         chunk_producer = loop.run_in_executor(
             ThreadPoolExecutor(max_workers=1, thread_name_prefix='chunk-producer'),
@@ -948,8 +967,9 @@ class Repository:
         name, tag = self._snapshot_digest_to_location_parts(digest)
         location = self.get_snapshot_location(name=name, tag=tag)
 
-        print(ef.bold + f'Uploading snapshot {name}' + ef.rs)
-        await self._as_coroutine(self.backend.upload, location, serialized_snapshot)
+        async with self._acquire_slot():
+            print(ef.bold + f'Uploading snapshot {name}' + ef.rs)
+            await self._awrap(self.backend.upload, location, serialized_snapshot)
 
         if state.bytes_reused:
             print(
@@ -1032,39 +1052,37 @@ class Repository:
                 except KeyError:
                     break
 
-                slot = await self._slots.get()
-                try:
-                    location = self._chunk_digest_to_location(digest)
+                location = self._chunk_digest_to_location(digest)
+
+                async with self._acquire_slot():
                     logger.info('Downloading %s', location)
-                    contents = await self._as_coroutine(self.backend.download, location)
+                    contents = await self._awrap(self.backend.download, location)
 
-                    logger.info('Decrypting %s', location)
-                    if self.props.encrypted:
-                        decrypted_contents = self.props.decrypt(
-                            contents,
-                            self.props.derive_shared_subkey(digest),
-                        )
-                    else:
-                        decrypted_contents = contents
-
-                    logger.info('Verifying %s', location)
-                    if self.props.hash_digest(decrypted_contents) != digest:
-                        raise exceptions.ReplicatError(
-                            f'Chunk at {location!r} is corrupted'
-                        )
-
-                    logger.info('Chunk %s referenced %d time(s)', location, len(refs))
-                    decrypted_view = memoryview(decrypted_contents)
-                    await asyncio.gather(
-                        *(
-                            loop.run_in_executor(
-                                executor, _write_chunk_ref, ref, decrypted_view
-                            )
-                            for ref in refs
-                        )
+                logger.info('Decrypting %s', location)
+                if self.props.encrypted:
+                    decrypted_contents = self.props.decrypt(
+                        contents,
+                        self.props.derive_shared_subkey(digest),
                     )
-                finally:
-                    self._slots.put_nowait(slot)
+                else:
+                    decrypted_contents = contents
+
+                logger.info('Verifying %s', location)
+                if self.props.hash_digest(decrypted_contents) != digest:
+                    raise exceptions.ReplicatError(
+                        f'Chunk at {location!r} is corrupted'
+                    )
+
+                logger.info('Chunk %s referenced %d time(s)', location, len(refs))
+                decrypted_view = memoryview(decrypted_contents)
+                await asyncio.gather(
+                    *(
+                        loop.run_in_executor(
+                            executor, _write_chunk_ref, ref, decrypted_view
+                        )
+                        for ref in refs
+                    )
+                )
 
         for snapshot_body in snapshots:
             snapshot_chunks = snapshot_body['chunks']
@@ -1166,22 +1184,16 @@ class Repository:
         )
 
         async def _delete_snapshot(location):
-            slot = await self._slots.get()
-            try:
-                await self._as_coroutine(self.backend.delete, location)
-                finished_snapshots_tracker.update()
-            finally:
-                self._slots.put_nowait(slot)
+            async with self._acquire_slot():
+                await self._awrap(self.backend.delete, location)
+            finished_snapshots_tracker.update()
 
         async def _delete_chunk(digest):
-            slot = await self._slots.get()
-            try:
-                await self._as_coroutine(
+            async with self._acquire_slot():
+                await self._awrap(
                     self.backend.delete, self._chunk_digest_to_location(digest)
                 )
-                finished_chunks_tracker.update()
-            finally:
-                self._slots.put_nowait(slot)
+            finished_chunks_tracker.update()
 
         with finished_snapshots_tracker:
             await asyncio.gather(*map(_delete_snapshot, snapshots_locations))
@@ -1202,9 +1214,7 @@ class Repository:
         to_delete = set()
 
         print(ef.bold + 'Fetching chunks list' + ef.rs)
-        for location in await self._as_coroutine(
-            self.backend.list_files, self.CHUNK_PREFIX
-        ):
+        async for location in self._aiter(self.backend.list_files, self.CHUNK_PREFIX):
             if location in referenced_locations:
                 logger.info('Chunk %s is referenced, skipping', location)
                 continue
@@ -1232,18 +1242,16 @@ class Repository:
         )
 
         async def _delete_chunk(location):
-            slot = await self._slots.get()
-            try:
-                await self._as_coroutine(self.backend.delete, location)
-                finished_tracker.update()
-            finally:
-                self._slots.put_nowait(slot)
+            async with self._acquire_slot():
+                await self._awrap(self.backend.delete, location)
+            finished_tracker.update()
 
         with finished_tracker:
             await asyncio.gather(*map(_delete_chunk, to_delete))
 
-        print(ef.bold + 'Running post-deletion cleanup' + ef.rs)
-        await self._as_coroutine(self.backend.clean)
+        async with self._acquire_slot():
+            print(ef.bold + 'Running post-deletion cleanup' + ef.rs)
+            await self._awrap(self.backend.clean)
 
     @utils.disable_gc
     def _benchmark_chunker(self, adapter, number=10, size=512_000_000):
@@ -1321,13 +1329,16 @@ class Repository:
                 os.path.commonpath([path, base_directory])
             ).as_posix()
             length = path.stat().st_size
-            slot = await self._slots.get()
-            try:
-                if skip_existing and await self._as_coroutine(
-                    self.backend.exists, name
-                ):
-                    logger.info('Skipping file %r (exists)', name)
-                else:
+            exists = False
+
+            if skip_existing:
+                async with self._acquire_slot():
+                    exists = await self._awrap(self.backend.exists, name)
+
+            if exists:
+                logger.info('Skipping file %r (exists)', name)
+            else:
+                async with self._acquire_slot() as slot:
                     logger.info('Uploading file %r', name)
                     stream = path.open('rb')
                     iowrapper = utils.tqdmio(
@@ -1338,16 +1349,13 @@ class Repository:
                         disable=self._quiet,
                         rate_limiter=rate_limiter,
                     )
-
                     with stream, iowrapper:
-                        await self._as_coroutine(
+                        await self._awrap(
                             self.backend.upload_stream, name, iowrapper, length
                         )
 
-                finished_tracker.update()
-                bytes_tracker.update(length)
-            finally:
-                self._slots.put_nowait(slot)
+            finished_tracker.update()
+            bytes_tracker.update(length)
 
         await asyncio.gather(*map(_upload_path, files))
         return utils.DefaultNamespace(files=files)
@@ -1358,4 +1366,4 @@ class Repository:
         except AttributeError:
             pass
 
-        await self._as_coroutine(self.backend.close)
+        await self._awrap(self.backend.close)
